/*
 * Licensed to the Apache Software Foundation (ASF) under one or more contributor license
 * agreements. See the NOTICE file distributed with this work for additional information regarding
 * copyright ownership. The ASF licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License. You may obtain a
 * copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable
 * law or agreed to in writing, software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License
 * for the specific language governing permissions and limitations under the License.
 */

package org.apache.myfaces.portlet.faces.application.view;

import java.beans.BeanInfo;
import java.io.IOException;

import javax.faces.application.Resource;
import javax.faces.application.ViewHandler;
import javax.faces.component.UIViewRoot;
import javax.faces.context.FacesContext;
import javax.faces.view.StateManagementStrategy;
import javax.faces.view.ViewDeclarationLanguage;
import javax.faces.view.ViewMetadata;

/**
 * This is a VDL that redirects everything to the primary unless there is a exception thrown in 
 * either buildView or renderview methods.  If this happens, it will delegate.  This is only used
 * with the "default" delegation model.
 * 
 * @param primary
 * @param secondary
 */
class AggregatingViewDeclatationLanguage extends ViewDeclarationLanguage
{
  private ViewDeclarationLanguage mPrimary;
  private ViewDeclarationLanguage mSecondary;
  
  private boolean mRenderSecondary = false;
  
 public AggregatingViewDeclatationLanguage(ViewDeclarationLanguage primary, ViewDeclarationLanguage secondary)
  {
    mPrimary = primary;
    mSecondary = secondary;
  }

  @Override
  public void buildView(FacesContext context, UIViewRoot root) throws IOException
  {
    if(!mRenderSecondary)
    {
      try
      {
        mPrimary.buildView(context, root);
        return;
      }
      catch (Throwable t)
      {
        mRenderSecondary=true;
      }
    }
    
    //There was an exception from the original renderView, I'm now
    //going to use the portal's renderer

    mSecondary.buildView(context, root);
  }

  @Override
  public UIViewRoot createView(FacesContext arg0, String arg1)
  {
    return mPrimary.createView(arg0, arg1);
  }

  @Override
  public BeanInfo getComponentMetadata(FacesContext arg0, Resource arg1)
  {
    return mPrimary.getComponentMetadata(arg0, arg1);
  }

  @Override
  public Resource getScriptComponentResource(FacesContext arg0, Resource arg1)
  {
    return mPrimary.getScriptComponentResource(arg0, arg1);
  }

  @Override
  public StateManagementStrategy getStateManagementStrategy(FacesContext arg0,
      String arg1)
  {
    return mPrimary.getStateManagementStrategy(arg0, arg1);
  }

  @Override
  public ViewMetadata getViewMetadata(FacesContext arg0, String arg1)
  {
    return mPrimary.getViewMetadata(arg0, arg1);
  }

  @Override
  public void renderView(FacesContext context, UIViewRoot root) throws IOException
  {
    if(!mRenderSecondary)
    {
      try
      {
        mPrimary.renderView(context, root);
        return;
      }
      catch (Throwable t)
      {
        mRenderSecondary = true;
      }
    }
    
    //There was an exception from the original renderView, I'm now
    //going to use the portal's renderer

    mSecondary.renderView(context, root);    
  }

  @Override
  public UIViewRoot restoreView(FacesContext arg0, String arg1)
  {
    UIViewRoot vr = mPrimary.restoreView(arg0, arg1);
    
    
    // Workaround Trinidad bug where it doesn't set the renderkit Id in the VR
    // after/during restore -- which causes a NPE in Mojarra 1.2.0.4
    if (vr != null && vr.getRenderKitId() == null)
    {
      ViewHandler outerViewHandler =
            arg0.getApplication().getViewHandler();
      vr.setRenderKitId(outerViewHandler.calculateRenderKitId(arg0));
    }
    return vr;
  }

}
