/*
 * Licensed to the Apache Software Foundation (ASF) under one or more contributor license
 * agreements. See the NOTICE file distributed with this work for additional information regarding
 * copyright ownership. The ASF licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License. You may obtain a
 * copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable
 * law or agreed to in writing, software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License
 * for the specific language governing permissions and limitations under the License.
 */

package org.apache.myfaces.portlet.faces.application.view;

import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.faces.application.StateManager;
import javax.faces.component.UIViewRoot;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;
import javax.faces.context.ResponseWriter;
import javax.portlet.MimeResponse;
import javax.portlet.RenderResponse;
import javax.portlet.ResourceResponse;
import javax.portlet.faces.Bridge;
import javax.portlet.faces.BridgeUtil;

import javax.portlet.faces.BridgeWriteBehindResponse;

import org.apache.myfaces.portlet.faces.bridge.BridgeImpl;
import org.apache.myfaces.portlet.faces.util.QueryString;
import org.apache.myfaces.shared_portlet.view.JspViewDeclarationLanguageBase;
import org.apache.myfaces.shared_portlet.view.ViewResponseWrapper;

public class PortletJspViewDeclarationLanguageImpl extends
    JspViewDeclarationLanguageBase
{
  private static final Logger sLOGGER = Logger.getLogger(PortletJspViewDeclarationLanguageImpl.class.getName());

  @Override
  protected void sendSourceNotFound(FacesContext context, String message)
  {
    // TODO Auto-generated method stub

  }

  @Override
  public void buildView(FacesContext context, UIViewRoot view) throws IOException
  {
    if(null == context)
    {
      //TODO: Localize Strings
      throw new NullPointerException("FacesContext is null");
    }
    
    if(null == view)
    {
      //TODO: Localize Strings
      throw new NullPointerException("UIViewRoot is null");
    }
    
    String viewId = view.getViewId();
    
    ExternalContext ec = context.getExternalContext();
    //TODO: Check to see if view is populated
    
    // update the JSTL locale attribute in request scope so that JSTL
    // picks up the locale from viewRoot. This attribute must be updated
    // before the JSTL setBundle tag is called because that is when the
    // new LocalizationContext object is created based on the locale.
    // Unfortunately we need to put this attribute on directly because the
    // JSTL Config.set API does not accommodate a portlet request.  Quite
    // possibly we may need an implementation adapter here but for now we
    // can hard code it.
    
    //TODO: MyFaces also sets the Locale on the response.  Do we need to do this here?
    ec.getRequestMap().put("javax.servlet.jsp.jstl.fmt.locale", context.getViewRoot().getLocale());
    
    if (sLOGGER.isLoggable(Level.FINE)) 
    {
        sLOGGER.fine("Before dispacthMessage to viewId " + viewId);
    }
    
    // save the original response
    MimeResponse originalResponse = (MimeResponse) ec.getResponse();

    // see whether the portlet wants to use a writeBehindResponse(Wrapper)
    Class<? extends BridgeWriteBehindResponse> writeBehindResponseWrapperClass = (Class<? extends BridgeWriteBehindResponse>) ec.getRequestMap().get(BridgeImpl.WRITE_BEHIND_RESPONSE);
    BridgeWriteBehindResponse wrapped = null;
    if (writeBehindResponseWrapperClass != null)
    {
      try
      {  
        wrapped = writeBehindResponseWrapperClass.newInstance();
        if ((BridgeUtil.getPortletRequestPhase() == Bridge.PortletPhase.RENDER_PHASE && wrapped instanceof RenderResponse) ||
            (BridgeUtil.getPortletRequestPhase() == Bridge.PortletPhase.RESOURCE_PHASE && wrapped instanceof ResourceResponse))                                                                                                                      
        {

          ec.setResponse(wrapped);
        }
        else
        {
          // warn and don't use
          wrapped = null;
          ec.log("Unable to use configured BridgeWriteBehindResponse instance because its not the appropriate PortletResponseWrapper for the associated phase.");
        }
        
      } catch (Exception e)
      {
        // Just warn and don't use;
        ec.log("Unable to instantiate configured BridgeWriteBehindResponse instance: ", e);
      }
    } 
    else
    {
      // Use the generic (non-writeBehindResponse) wrapper 
      if (BridgeUtil.getPortletRequestPhase() == Bridge.PortletPhase.RENDER_PHASE)
      {
        wrapped = new BridgeViewHandlerRenderResponseWrapper();
      }
      else
      {
        wrapped = new BridgeViewHandlerResourceResponseWrapper();
      }
      ec.setResponse(wrapped);
      // don't set write behind as this one doesn't support the Faces specific implementation dependent interfaces
    }
    // set request attribute indicating we can deal with content
    // that is supposed to be delayed until after JSF tree is ouput.
    ec.getRequestMap().put(Bridge.RENDER_CONTENT_AFTER_VIEW, Boolean.TRUE);


    // build the view by executing the page
    ec.dispatch(viewId);

    if (wrapped != null)
    {
      // replace the original response
      ec.setResponse(originalResponse);
      
      // Check whether our wrapper participated in the the writeBehind mechanism and/or
      // something else already did (servlet filter)
      Object o = ec.getRequestMap().remove(Bridge.AFTER_VIEW_CONTENT);
      
      if (o == null && wrapped.hasFacesWriteBehindMarkup())
      {
          o = (wrapped.isChars()) ? (Object) wrapped.getChars() : (Object) wrapped.getBytes();
      }
        
      //This is for backward compatibility with the Portlet 1.0 bridge's filter
      //It simply replaces the "wrapped" object we got from parsing the JSP.
      if(null != o)
      {
        ViewResponseWrapper afterContentWrapper = new AfterViewContentResponseWrapper(o);
        
        // Now set it so it will be written after.
        setAfterViewTagResponseWrapper(ec, afterContentWrapper);
      }
      else
      {
        // write ahead -- just write this to the restored response
        wrapped.flushMarkupToWrappedResponse();
      }
    }
  }
  
  @Override
  protected void flushBufferToWriter(StringBuffer buff, Writer writer) throws IOException
  {
    FacesContext facesContext = FacesContext.getCurrentInstance();
    StateManager stateManager = facesContext.getApplication().getStateManager();

    StringWriter stateWriter = new StringWriter();
    ResponseWriter realWriter = facesContext.getResponseWriter();
    facesContext.setResponseWriter(realWriter.cloneWithWriter(stateWriter));

    Object serializedView = stateManager.saveView(facesContext);

    stateManager.writeState(facesContext, serializedView);
    facesContext.setResponseWriter(realWriter);

    String state = stateWriter.getBuffer().toString();

    String marker = getStateMarker(buff);
    if(marker != null)
    {
      int markerLen = marker.length();
    
      int lastFormMarkerPos = 0;
      int formMarkerPos = 0;
      
      // Find all state markers and write out actual state instead
      while ((formMarkerPos = buff.indexOf(marker, lastFormMarkerPos)) > -1)
      {
        // Write content before state marker
        writePartialBuffer(buff, lastFormMarkerPos, formMarkerPos, writer);
        // Write state and move position in buffer after marker
        writer.write(state);
        formMarkerPos += markerLen;
        lastFormMarkerPos = formMarkerPos;
      }
      
      // Write content after last state marker
      if (lastFormMarkerPos < buff.length())
      {
        writePartialBuffer(buff, lastFormMarkerPos, buff.length(), writer);
      }
    }
    else
    {
      writer.write(buff.toString());
    }
  }
  
  protected boolean actuallyRenderView(FacesContext facesContext, UIViewRoot viewToRender)
      throws IOException
  {
      if (super.actuallyRenderView(facesContext, viewToRender))
      {
        // Check again to see if the processing of the Faces view triggered a redirect
        QueryString redirectParams = (QueryString) facesContext.getExternalContext()
                          .getRequestMap().get(BridgeImpl.REDIRECT_VIEWPARAMS);
        if ((redirectParams != null))
        {
          // set a flag indicating we are in a renderRedirect so the ViewHandler will know to dispatch.include instead of forward
          // mark it as such so the next dispatch will use an include
          facesContext.getExternalContext().getRequestMap().put(BridgeImpl.HAS_RENDER_REDIRECTED_AFTER_FORWARD, Boolean.TRUE);
          return false;
        }
        
        // otherwise -- view was rendered -- check to see if the view rendering created any after view content manually
        // I.e. JSF EL wrote content directly into an after view buffer and placed on the bridge's request attr.  This
        // check supports the TCK test.
        Object o = facesContext.getExternalContext().getRequestMap().remove(Bridge.AFTER_VIEW_CONTENT);
                 
        //This is for backward compatibility with the Portlet 1.0 bridge's filter
        //It simply replaces the "wrapped" object we got from parsing the JSP.
        if(null != o)
        {
          ViewResponseWrapper afterContentWrapper = new AfterViewContentResponseWrapper(o);
          
          // Now set it so it will be written after.
          setAfterViewTagResponseWrapper(facesContext.getExternalContext(), afterContentWrapper);
        }
        return true;   
      }
      else
      {
        return false;
      }

  }
  
  /**
   * Return a form state marker if there is one or null if one could
   * no be found.
   * @param buff
   * @return
   */
  private String getStateMarker(StringBuffer buff)
  {
    if(null == mStateMarker)
    {
      if(buff.indexOf(MYFACES_FORM_STATE_MARKER) > -1)
      {
        mStateMarker = MYFACES_FORM_STATE_MARKER;
      }
      else if(buff.indexOf(RI_FORM_STATE_MARKER) > -1)
      {
        mStateMarker = RI_FORM_STATE_MARKER;
      }
      
      //LOG A WARNING HERE
    }
    
    return mStateMarker;
  }
  
  private class AfterViewContentResponseWrapper implements ViewResponseWrapper
  {
    Object buffer;

    public AfterViewContentResponseWrapper(Object o)
    {
      buffer = o;
    }
    
    public void flushToWriter(Writer writer, String encoding)
        throws IOException
    {
      if(buffer instanceof char[])
      {
        writer.write((char[]) buffer);
      }
      else
      {
        writer.write(new String((byte[]) buffer, encoding));
      }
      writer.flush();
    }
  };
  
  private String mStateMarker;
  private static final String MYFACES_FORM_STATE_MARKER = "<!--@@JSF_FORM_STATE_MARKER@@-->";
  private static final String RI_FORM_STATE_MARKER   = "~com.sun.faces.saveStateFieldMarker~";

}
