/*
 * Licensed to the Apache Software Foundation (ASF) under one or more contributor license
 * agreements. See the NOTICE file distributed with this work for additional information regarding
 * copyright ownership. The ASF licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License. You may obtain a
 * copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable
 * law or agreed to in writing, software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License
 * for the specific language governing permissions and limitations under the License.
 */

package org.apache.myfaces.portlet.faces.context;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.el.ELContext;
import javax.el.ELContextEvent;
import javax.el.ELContextListener;

import javax.faces.FacesException;
import javax.faces.application.Application;
import javax.faces.component.UIViewRoot;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;
import javax.faces.context.FacesContextFactory;
import javax.faces.context.FacesContextWrapper;
import javax.faces.lifecycle.Lifecycle;
import javax.portlet.PortletResponse;
import javax.portlet.faces.Bridge;
import javax.portlet.faces.annotation.PortletNamingContainer;

import org.apache.myfaces.portlet.faces.bridge.BridgeImpl;
import org.apache.myfaces.portlet.faces.el.PortletELContextImpl;
import org.apache.myfaces.shared_portlet.util.ExternalContextUtils;

public class PortletFacesContextFactoryImpl
  extends FacesContextFactory
{
  private static final Logger sLOG= Logger.getLogger(PortletFacesContextFactoryImpl.class.getName());
  private FacesContextFactory mFactory;
  private ELContext mElContext;
  
  public PortletFacesContextFactoryImpl(FacesContextFactory factory)
  {
    mFactory = factory;
  }
  
  @Override
  public FacesContextFactory getWrapped()
  {
    return mFactory;
  }

  public FacesContext getFacesContext(Object context, Object request, Object response, Lifecycle lifecycle)
    throws FacesException
  {
    // if in portlet environment -- do a portlet container neutral test
    // first in case we are packaged in a web app that isn't deployed 
    // on a portlet container/as a portlet. Note:  can't use the BridgeUtil
    // method as that call requires the facesContext to exist.
    FacesContext fc = getWrapped().getFacesContext(context, request, response, lifecycle);
      
    if(ExternalContextUtils.isPortlet(fc.getExternalContext()))
    {
      fc = new FacesContextImpl(fc);
    }
      
    return fc;
  }

  private class FacesContextImpl extends FacesContextWrapper
  {
    private  FacesContext mWrappedContext;
    
    public FacesContextImpl(FacesContext context)
    {
      mWrappedContext = context;
      setCurrentInstance(this);
    }

    public FacesContext getWrapped()
    {
      return mWrappedContext;
    }
    
    /**
     * <p>
     * Return the <code>ELContext</code> instance for this <code>FacesContext</code> instance.
     * This <code>ELContext</code> instance has the same lifetime and scope as the
     * <code>FacesContext</code> instance with which it is associated, and may be created lazily the
     * first time this method is called for a given <code>FacesContext</code> instance. Upon
     * creation of the ELContext instance, the implementation must take the following action:
     * </p>
     * 
     * <ul>
     * 
     * <li>
     * <p>
     * Call the {@link ELContext#putContext} method on the instance, passing in
     * <code>FacesContext.class</code> and the <code>this</code> reference for the
     * <code>FacesContext</code> instance itself.
     * </p>
     * </li>
     * 
     * <li>
     * <p>
     * If the <code>Collection</code> returned by {@link
     * javax.faces.application.Application#getELContextListeners} is non-empty, create an instance of
     * {@link javax.el.ELContextEvent} and pass it to each {@link javax.el.ELContextListener} instance
     * in the <code>Collection</code> by calling the {@link
     * javax.el.ELContextListener#contextCreated} method.
     * </p>
     * </li>
     * 
     * </ul>
     * 
     * <p>
     * The default implementation throws <code>UnsupportedOperationException</code> and is provided
     * for the sole purpose of not breaking existing applications that extend this class.
     * </p>
     * 
     * @throws IllegalStateException
     *           if this method is called after this instance has been released
     * 
     * @since 1.2
     */
    @Override
    public ELContext getELContext()
    {
      if (mElContext == null)
      {
        Application app = getApplication();
        mElContext = new PortletELContextImpl(app.getELResolver());
        // Use one set as current instance in case we are wrapped
        mElContext.putContext(FacesContext.class, FacesContext.getCurrentInstance());
        UIViewRoot root = getViewRoot();
        if (null != root)
        {
          mElContext.setLocale(root.getLocale());
        }
        
        // Now notify any listeners that we have created this context
        ELContextListener[] listeners = app.getELContextListeners();
        if (listeners.length > 0)
        {
          ELContextEvent event = new ELContextEvent(mElContext);
          for (ELContextListener listener:listeners)
          {
            listener.contextCreated(event);
          }
        }
      }
      return mElContext;
    }

    @Override
    public void setViewRoot(UIViewRoot root)
    {
      FacesContext fc = getWrapped();
      fc.setViewRoot(root);
      
      // if ViewRoot annotated with PortletNamingContainer
      // it supports portlet namespacing content -- mark
      // response so consumers can detect
      if (fc.getViewRoot().getClass().getAnnotation(PortletNamingContainer.class) != null)
      {
        try
        {
          PortletResponse pr = (PortletResponse) fc.getExternalContext().getResponse();
          pr.addProperty(Bridge.PORTLET_NAMESPACED_RESPONSE_PROPERTY, "true");
        }
        catch (Exception e)
        {
          sLOG.log(Level.WARNING, "Response does not appear to be a valid type", e);
        }
      }
    }

    @Override
    public void release()
    {
      mElContext = null;
      // Reset attributes to those that existed before we were called
      // This is done because some (local) portals run/render all portlets in the same request
      ExternalContext ec = getWrapped().getExternalContext();
      Map<String,Object> reqAttrs = ec.getRequestMap();
      
      @SuppressWarnings("unchecked")
      List<String> preExistingAttrs = (List<String>)reqAttrs.get(BridgeImpl.PREEXISTING_ATTRIBUTE_NAMES);
      if (preExistingAttrs != null)
      {
        ArrayList<String> removeList = (ArrayList<String>) new ArrayList(preExistingAttrs.size());
  
  
        Set<String> keys= reqAttrs.keySet();
        
        for(String key:keys)
        {
          if (!preExistingAttrs.contains(key))
          {
            // Postpone the remove until after the iteration as it causes a ConcurrentModificationException on some appServers (WebSphere)
            removeList.add(key);
          }
        }
        
        // Postpone the remove until after the iteration as it causes a ConcurrentModificationException on some appServers (WebSphere)
        for(Iterator<String> iter = removeList.iterator(); iter.hasNext();)
        {
          reqAttrs.remove(iter.next());
        }
      }
      
      if(ec instanceof PortletExternalContextImpl)
      {
        ((PortletExternalContextImpl)ec).release();
      }

      super.release();
    }
  }
}
